/*
 * (c) 2008-2009 Technische Universität Dresden
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */
#include "ldso_backend.h"
#include "debug.h"
#include "globals.h"

#include <l4/cxx/iostream.h>
#include <l4/cxx/l4iostream.h>

#include <l4/re/env>
#include <l4/re/rm>
#include <l4/re/dataspace>
#include <l4/re/debug>

#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <unistd.h>

using L4Re::Rm;

enum
{
  PROT_NONE  = 0,
  PROT_READ  = 1,
  PROT_WRITE = 2,
  PROT_EXEC  = 4,
};

enum
{
  MAP_PRIVATE    = 2,
  MAP_SHARED     = 1,
  MAP_ANONYMOUS  = 32,
  MAP_FIXED      = 16,
  MAP_EXECUTABLE = 0, // ignored
  MAP_DENYWRITE  = 0, // ignored
};

#define MAP_FAILED ((void*)~0UL)

struct stat
{
  long st_size;
  int st_mode;
  unsigned long st_dev;
  int st_ino;
};

struct Dl_syscall_map
{
  void *(*mmap)(void *, size_t, int, int, int, off_t);
  int (*munmap)(void *, size_t);
  int (*mprotect)(void *, size_t, int);
  int (*open)(char const *, int, int);
  int (*close)(int);
  int (*read)(int, void *, size_t);
  int (*write)(int, char const *, size_t);
  int (*fstat)(int, struct stat *);
};

struct Dl_env_ops
{
  void *(*malloc)(size_t);
  void (*free)(void*);

  unsigned long (*cap_alloc)();
  void (*cap_free)(unsigned long);
};

struct File
{
  L4::Cap<L4Re::Dataspace> ds;
  l4_umword_t size;
  l4_addr_t ptr;
  char const *addr;
};

enum
{
  MAX_FILES = 20,
  FD_OFFSET = 8,
};

static int __fd_next = 0;
static File __files[MAX_FILES];

static int
alloc_fd()
{
  if (__fd_next != -1)
    {
      int fd = __fd_next;
      __fd_next = -1;
      return fd;
    }

  for (int i = 0; i < MAX_FILES; ++i)
    {
      if (!__files[i].ds.is_valid())
	return i;
    }

  return -1;
}

static void
free_fd(int fd)
{
  __fd_next = fd;
  __files[fd].ds = L4::Cap<L4Re::Dataspace>();
}


static Dbg dbg(Dbg::POSIX_be, "POSIXbe");

static
bool
__release(L4::Cap<L4Re::Dataspace> const &c)
{
  c->release();
  int l = c.validate(L4Re::This_task).label();
  //printf("release_cap: %lx %d\n", c.cap(), l);
  if (!l)
    {
      Global::cap_alloc.free(c);
      return true;
    }

  return false;
}

static int
_close(int fd)
{
  dbg.printf("close(fd=%d)", fd);
  fd -= FD_OFFSET;

  if (fd >= MAX_FILES)
    {
      dbg.cprintf("=-1 (invalid fd)\n");
      return -1;
    }

  if (!__files[fd].ds.is_valid())
    {
      dbg.cprintf("=-1 (invalid fd)\n");
      return -1;
    }

  dbg.cprintf("{cap=[C:%lx]", __files[fd].ds.cap());

  if (__files[fd].addr)
    {
      dbg.cprintf(" detach file (addr=%lx)", l4_umword_t(__files[fd].addr));
      L4Re::Env::env()->rm()->detach(l4_addr_t(__files[fd].addr), 0);
    }

  __release(__files[fd].ds);

  free_fd(fd);
  dbg.cprintf("}=0\n");

  return 0;
}

static int
_open(char const *path, int flags, int mode)
{
  dbg.printf("open(path='%s', flags=%d, mode=0%o)", path, flags, mode);

  L4::Cap<L4Re::Dataspace> file = Global::cap_alloc.alloc<L4Re::Dataspace>();

  if (!file.is_valid())
    {
      dbg.cprintf("=-1 (could not allocate cap)\n");
      return -1;
    }

  long err = L4Re::Env::env()->names()->query(path, file);

  if (err < 0)
    {
      Global::cap_alloc.free(file, L4Re::This_task);
      dbg.cprintf("=-1 (could not find file)\n");
      return -1;
    }

  int fd = alloc_fd();
  if (fd < 0)
    {
      Global::cap_alloc.free(file, L4Re::This_task);
      dbg.cprintf("=-1 (out of fd-s)\n");
      return -1;
    }


  __files[fd].ds = file;
  // FIXME: dont know if this is right
  __files[fd].ds->take();

  __files[fd].addr = 0;
  __files[fd].ptr = 0;
  __files[fd].size = __files[fd].ds->size();

  dbg.cprintf("=%d {cap=[C:%lx]}\n", fd + FD_OFFSET, __files[fd].ds.cap());
  return fd + FD_OFFSET;
}

static int
_fstat(int fd, struct stat *buf)
{
  static int fake = 0;
  dbg.printf("fstats(fd=%d, buf=%p)", fd, buf); 
  fd -= FD_OFFSET;

  if (fd<0 || fd >= MAX_FILES)
    {
      dbg.cprintf("=-1 (invalid fd)\n");
      return -1;
    }
  

  buf->st_size = __files[fd].size;
  buf->st_mode = 0666;
  buf->st_dev = __files[fd].ds.cap();
  buf->st_ino = ++fake;
  dbg.cprintf("=0 {size=0x%lx, ino=0x%x, mode=0%o}\n", __files[fd].size,
      buf->st_ino, buf->st_mode);
  return 0;
}

#if 0
struct L_region
{
  L4::Capability c;
  l4_addr_t addr;
  unsigned long size;
  unsigned long offs;
  bool ro, ds;

  L_region(L4::Capability const &c, l4_addr_t a, unsigned long sz,
           unsigned long offs, bool ro, bool ds)
    : c(c), addr(a), size(sz), offs(offs), ds(ds) {}
};
#endif 

static int
_munmap(void *start, size_t len)
{
  using namespace L4;
  using namespace L4Re;
  dbg.printf("munmap(start=%p, len=0x%lx)=0\n", start, (unsigned long)len);

  int err;
  Cap<Dataspace> ds;

  while (1)
    {
      err = Env::env()->rm()->detach(l4_addr_t(start), len, &ds, This_task);
      if (err < 0)
	return err;

      switch (err & Rm::Detach_result_mask)
	{
	case Rm::Split_ds:
	  if (ds.is_valid())
	    ds->take();
	  return 0;
	case Rm::Detached_ds:
	  if (ds.is_valid())
	    __release(ds);
	  break;
	default:
	  break;
	}

      if (!(err & Rm::Detach_again))
	return 0;
    }
}

static void *
_mmap(void *start, size_t len, int prot, int flags, int fd, off_t offset)
{
  using namespace L4Re;

  start = (void*)l4_trunc_page(l4_addr_t(start));
  len   = l4_round_page(len);
#if 0
  // dumb hack
  flags |= MAP_PRIVATE;
  prot  |= PROT_WRITE;
  // end: hack
#endif

  dbg.printf("mmap(s=%p, len=0x%lx, p=0x%x, f=0x%x, fd=%d, offs=0x%lx)",
         start, (unsigned long)len, prot, flags, fd, (unsigned long)offset);
  L4::Cap<L4Re::Dataspace> ds = Global::cap_alloc.alloc<L4Re::Dataspace>();
  if (!ds.is_valid())
    {
      dbg.cprintf("=-1 (out of caps)\n");
      return MAP_FAILED;
    }

  l4_umword_t size = (len + L4_PAGESIZE-1) & ~(L4_PAGESIZE-1);

  if (flags & (MAP_ANONYMOUS | MAP_PRIVATE))
    {
      Global::allocator->alloc(size, ds); 

      if (!ds.is_valid())
	{
	  dbg.cprintf("=-1 (out of memory)\n");
	  return MAP_FAILED;
	}
    }

  if (!(flags & MAP_ANONYMOUS))
    {
      fd -= FD_OFFSET;
      if (fd < 0 || fd >= MAX_FILES)
	{
	  dbg.cprintf("=-1 (invalid fd)\n");
	  return MAP_FAILED;
	}

      if (!__files[fd].ds.is_valid())
	{
	  dbg.cprintf("=-1 (invalid fd)\n");
	  return MAP_FAILED;
	}

      if (size + offset > l4_round_page(__files[fd].size))
	{
	  dbg.cprintf("=-1 (offset too big)\n");
	  return MAP_FAILED;
	}

      if (flags & MAP_PRIVATE)
	{
	  ds->copy_in(0, __files[fd].ds, l4_trunc_page(offset),
	      l4_round_page(size));
	  offset = 0;
	}
      else
	{
          ds = __files[fd].ds;
	  ds->take();
	}
    }

  if (!(flags & MAP_FIXED) && start == 0)
    start = (void*)L4_PAGESIZE;

  int err;
  unsigned rm_flags = 0;
  char *data = (char *)start;
  L4::Cap<Rm> r = Env::env()->rm();
  l4_addr_t overmap_area;

  if (flags & MAP_FIXED)
    {
      overmap_area = l4_addr_t(start);
      err = r->reserve_area(&overmap_area, size);
      if (err < 0)
	overmap_area = L4_INVALID_ADDR;
      else
	rm_flags |= Rm::In_area;

      while (1)
	{
	  L4::Cap<L4Re::Dataspace> ds;
	  err = r->detach(l4_addr_t(start), len, &ds, This_task);
	  if (err == -L4_ENOENT)
	    break;

	  if (err < 0)
	    return MAP_FAILED;

	  switch (err & Rm::Detach_result_mask)
	    {
	    case Rm::Split_ds:
	      if (ds.is_valid())
		ds->take();
	      break;
	    case Rm::Detached_ds:
	      if (ds.is_valid())
		__release(ds);
	      break;
	    default:
	      break;
	    }

	  if (!(err & Rm::Detach_again))
	    break;
	}
    }

  if (!(flags & MAP_FIXED))  rm_flags |= Rm::Search_addr;
  if (!(prot & PROT_WRITE))  rm_flags |= Rm::Read_only;

  err = r->attach(&data, size, rm_flags, ds, offset);

  if (rm_flags == Rm::In_area)
    r->free_area(overmap_area);

  if (err < 0)
    return MAP_FAILED;

  dbg.cprintf("=0 {mapped to %p-%p}\n", data, data + size);

  if (start && !data)
    return MAP_FAILED;

  return data;
}

int _mprotect(void *a, size_t sz, int prot)
{
  dbg.printf("mprotect(%p, sz=0x%lx, p=0x%x)=0\n", a, (unsigned long)sz, prot);
  return (prot & PROT_WRITE) ? -1 : 0;
}

static int
_read(int fd, void *buf, size_t cnt)
{
  fd -= FD_OFFSET;
  if (fd < 0 || fd >= MAX_FILES)
    return -1;

  if (!__files[fd].ds.is_valid())
    return -1;

  if (!__files[fd].addr)
    {
      void const *file = (void*)L4_PAGESIZE;
      long err = L4Re::Env::env()->rm()->attach(&file, __files[fd].size,
	Rm::Search_addr | Rm::Read_only, __files[fd].ds, 0);

      if (err < 0)
	return err;

      if (file)
	__files[fd].addr = (char const*)file;
    }

  long l = cnt;
  if ((long)__files[fd].size - (long)__files[fd].ptr < l)
    l = __files[fd].size - __files[fd].ptr;

  if (l > 0)
    {
      memcpy(buf, __files[fd].addr + __files[fd].ptr, l);
      __files[fd].ptr += l;
      return l;
    }
  return 0;
}

static int
_write(int fd, char const *str, size_t len)
{
  dbg.printf("write(fd=%d, str=%p. len=%lx)\n", fd, str, (unsigned long)len);
  if (fd <= 0 || fd > 2)
    {
      Dbg(Dbg::Warn, "POSIXbe").printf("write access to read only file (fd=%d)\n", fd);
      return -1;
    }

  return write(fd, str, len);
}

Dl_syscall_map __ldso_dl_syscall_map = 
{
  _mmap,
  _munmap,
  _mprotect,
  _open,
  _close,
  _read,
  _write,
  _fstat
};
