/*
 * (c) 2008-2009 Technische Universität Dresden
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU Lesser General Public License 2.1.
 * Please see the COPYING-LGPL-2.1 file for details.
 */
#include <l4/libloader/script>
#include <l4/libloader/stack>
#include <l4/re/util/debug>

#include <l4/cxx/iostream.h>

#include <cctype>
#include <cstring>

using L4Re::Util::Err;

bool
Ldr::Scanner::check_leading_space(int min_space, String::Index start)
{
  String::Index p = start;
  int i;
  for (i = 0; !_text.eof(p) && *p && *p != '\n' && isspace(*p); ++i, ++p)
    ;

  if (!_text.eof(p))
    {
      if (*p == '#')
	return true;
      if (*p == '\n')
	return true;
    }

  return i >= min_space;
}

Ldr::Scanner
Ldr::Scanner::block(int min_space, int skip_lines)
{
  Scanner n(*this);
  for (;skip_lines; --skip_lines)
    _next_line();

  String::Index p = _cursor;
  while (check_leading_space(min_space, p))
    {
      while (!_text.eof(p) && *p && *p != '\n')
	++p;

      if (!_text.eof(p) && *p == '\n')
	{
	  ++_ctxt;
	  ++p;
	}
    }
  
  n._text = _text.substr(_cursor, p - _cursor);
  n._cursor = n._text.start();

  _cursor = p;
  return n;
}

Ldr::Token
Ldr::Scanner::get_operator(String::Index e)
{
  switch (*e)
    {
    case '=': return Token(Token::Assign, String(e, 1));
    case '(': return Token(Token::L_paren, String(e, 1));
    case ')': return Token(Token::R_paren, String(e, 1));
    case ',': return Token(Token::Comma, String(e, 1));
    case ';': return Token(Token::Semicolon, String(e, 1));
    default: break;
    }

  if (!_text.eof(e+1) && *e == '-' && e[1] == '>')
    return Token(Token::Pointsat, String(e, 2));

  return Token();
}

cxx::String
Ldr::Scanner::read_string(String::Index cur)
{
  String::Index e = cur;
  while (!_text.eof(e))
    {
      if (*e == '\\')
	{
	  if (_text.eof(e+1))
	    return String();

	  e += 2;
	  continue;
	}

      if (*e == '"')
	return _text.substr(cur, e - cur);

      ++e;
    }

  return String();
}

Ldr::Token
Ldr::Scanner::next_token()
{
  String::Index e = _cursor;
  while (1)
    {
      while (!_text.eof(e) && isspace(*e))
	{
	  if (*e == '\n')
	    ++_ctxt;
	  ++e;
	}

      if (!_text.eof(e) && *e == '#')
	{
	  while (!_text.eof(e) && *e != '\n')
	    ++e;

	  if (!_text.eof(e))
	    {
	      ++_ctxt; ++e;
	    }
	}
      else
	break;
    }


  if (_text.eof(e))
    return Token();

  if (*e == '"')
    {
      cxx::String s = read_string(e+1);
      if (!s.empty())
	_cursor = s.start() + s.len() + 1;
      else
	++_cursor;

      return Token(Token::String, s);
    }

    {
      Token op = get_operator(e);
      if (op.type() != Token::Eof)
	{
	  _cursor = op.text().start() + op.text().len();
	  return op;
	}
    }

  if (*e >= '0' && *e <= '9')
    {
      String::Index k = e;
      int base = 10;
      int digit = 0;
      long n = 0;
      ++k;
      if (*e == '0')
	{
	  if (*k == 'x' || *k == 'X')
	    {
	      base = 16;
	      ++k;
	    }
	  else
	    base = 8;
	}
      else
	n = *e - '0';

      while (!_text.eof(k))
	{
	  digit = 200;
	  if (*k >= '0' && *k <= '9')
	    digit = *k - '0';
	  else if (*k >= 'a' && *k <= 'f')
	    digit = *k - 'a' + 10;
	  else if (*k >= 'A' && *k <= 'F')
	    digit = *k - 'A' + 10;

	  if (digit >= base)
	    break;

	  n = n * base + digit;
	  ++k;
	}
      _cursor = k;
      return Token(Token::Number, _text.substr(e, k-e), n);
    }

  String::Index k = e;
  while (!_text.eof(k))
    {
      if (isspace(*k) || get_operator(k).type() != Token::Eof)
	{
	  _cursor = k;
	  return Token(Token::Name, _text.substr(e, k - e));
	}

      ++k;
    }

  return Token();
}

#if 0
static void
__print_eol(char const *c)
{
  while (*c && *c != '\n')
    outchar(*c++);

  outchar('\n');
}
#endif

#if 0
int
Ldr::Script::open_file(char const *file)
{
  Ldr::Dataspace *f = Boot_fs::open_file(file);
  if (!f)
    return -L4_ENOENT;

  Ldr::Dataspace_cont *fc = dynamic_cast<Ldr::Dataspace_cont*>(f);

  if (!fc)
    {
      _err.printf("script parser supports only contiguous data spaces\n");
      return -L4_ENOENT;
    }

  _file   = file;
  _text = String(f->address(0).adr<char const*>(), fc->size());
  rewind_stream();
  return 0;
}
#endif

bool
Ldr::Scanner::_is_quote(char const *c) const
{
  return (*c == '"') || (*c == '\'');
}

char const *
Ldr::Scanner::_parse_quoted(char const *c) const
{
  char q = *c;
  ++c;
  for (char l = 0; !eof() && (*c != q || l == '\\'); l = *c++)
    ;

  return c;
}

void
Ldr::Scanner::_skip_space()
{
  char l = 0;
  for (;;++_cursor)
    {
      if (eof())
	return;

      if (l == '\\')
	continue;

      if (*_cursor == '\n')
	++_ctxt;
      else if (!isspace(*_cursor))
	return;
    }
}

bool
Ldr::Scanner::_is_comment(char const *c) const
{
  return *c == '#';
}

void
Ldr::Scanner::_next_line()
{
  for (; !eof() && *_cursor != '\n'; ++_cursor)
    ;

  ++_ctxt;
  ++_cursor;
}


bool
Ldr::Script::next_entry()
{
  _dbg.printf("%s:%d: (%p-%p) goto next entry (%p)\n", file(), line(),
             _text.start(), _text.end(), _cursor);
  while (!eof())
    {
      if (*_cursor == '!' || *_cursor == '[' || *_cursor == '\n')
	_next_line();
      else if (isspace(*_cursor))
	_next_line();
      else if (_is_comment(_cursor))
	_next_line();
      else
	break;
    }

  _dbg.printf("%s: now on line %d\n", file(), line());

  return !eof();
}


char const *
Ldr::Script::_push_arg(Stack *st, char const *s, char const *e)
{
  char const *_s = st->push((char)0);
  char _q = 0;
  while (e > s)
    {
      --e;

      if (!_is_quote(e))
	_s = st->push(*e);
      else if (_q != 0 && *e != _q)
	_s = st->push(*e);
      else if (_q != 0 && *e == _q && e > s && e[-1] == '\\')
	{
	  _s = st->push(*e);
	  --e;
	}
      else if (_q != 0 && *e == _q)
	_q = 0;
      else if (_q == 0)
	_q = *e;
    }
  return _s;
}

bool
Ldr::Script::_match_section(String const &l, char const *sect, String *name)
{
  String _l = l.substr(1);
  String::Index i = _l.find(":]");
  if (_l.eof(i))
    {
      _err.printf("%s:%d: syntax error (missing ']')\n", file(), line());
      return false;
    }

  if (strncmp(_l.start(), sect, i - _l.start()) != 0)
    return false;

  _dbg.printf("%s:%d: found section '%s'\n", file(), line(), sect);

  if (*i == ':' && name)
    {
      *name = _l.substr(i+1);
      String::Index e = name->find("]");
      if (name->eof(e))
	{
	  _err.printf("%s:%d: syntax error (missing ']')\n", file(), line());
	  return false;
	}
      *name = name->head(e);

      _dbg.printf("%s:%d:   section '%s' instance name '%.*s'\n", file(),
	          line(), sect, name->len(), name->start());
    }
  else if (name)
    *name = String();

  return true;
}

Ldr::String
Ldr::Script::tail() const
{
  if (!eof(_cursor))
    return _text.substr(_cursor);

  return String();
}

Ldr::String
Ldr::Script::read_line()
{
  if (eof())
    return String(_cursor, _cursor);

  String t = tail();
  String::Index i = t.find("\n", _cursor);
  return t.head(i);
}


bool
Ldr::Script::goto_section(char const *sect, String *name)
{
  while (1)
    {
      String l = read_line();
      if (eof(l))
	  return false;

      if (l[0] == '[')
	{
	  if (_match_section(l, sect, name))
	    {
	      _next_line();
	      return true;
	    }
	}

      _next_line();
    }
}

void
Ldr::Script::skip_entry()
{ _next_line(); }

bool
Ldr::Script::parse_entry(Stack *stack,
    char const **argv0, char const **argvl)
{
  char const *current;

  if (eof() || isspace(*_cursor))
    return false;

  /* Note, there's no support for escaping quotes! */

  while (!eof() && *_cursor != '\n')
    {
      current = _cursor;

      /* Start of a quote? */
      if (_is_quote(_cursor))
	{
	  _cursor = _parse_quoted(_cursor);
	  _cursor++;
	}

      /* Forward to next whitespace */
      while (!eof() && !isspace(*_cursor) && !_is_comment(_cursor))
	_cursor++;

      if (_cursor > current)
	{
	  *argvl = _push_arg(stack, current, _cursor);
	  if (!*argv0)
	    *argv0 = *argvl;
	}

      if (!eof() && _is_comment(_cursor))
	{
	  _next_line();
	  return true;
	}

      if (*_cursor == '\n')
	{
	  _next_line();
	  return true;
	}

      if (!eof())
	_cursor++;
    }
  return true;
}

