/*
 * (c) 2008-2009 Technische Universität Dresden
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */
#include <l4/libloader/script>
#include "debug.h"
#include "name_space.h"
#include "obj_reg.h"
#include "config_namespace.h"
#include "script_invoke.h"

#include <l4/re/service>
#include <l4/libloader/name_space_cfg>

#include <cstring>
#include <cstdlib>

static Dbg info(Dbg::Name_space, "ns");
static Ldr::Name_space ns_ns;
Ldr::Name_space *Ldr::local_root_ns()
{ return &ns_ns; }

bool operator == (cxx::String const &a, char const *b)
{
  if ((int)strlen(b) != a.len())
    return false;

  return strncmp(a.start(), b, a.len()) == 0;
}

namespace Ldr {
namespace Names {
  using namespace L4Re::Util::Names;
}
using Ldr::Name_space;

static
Names::Name  namedup(Names::Name const &n)
{
  char *ne = (char*)malloc(n.len());
  if (!ne)
    return Names::Name(0,0);
  memcpy(ne, n.start(), n.len());
  return Names::Name(ne, n.len());
}

static
void freename(Names::Name const &n)
{
  free(const_cast<char*>(n.start()));
}


class Ldr_names_parser : public Names_parser
{
private:
  Ldr_obj_mgr mgr;
public:
  Ldr_names_parser(Ldr::Scanner *scan, Name_space *ns)
  : Names_parser(scan, ns, &mgr) {}

  void check_invocation(Val const &obj)
  {
    if (obj.is_local_entry())
      throw Parse_exception(_scan, Ldr::Token(),
                            "cannot invoke locally registered objects");
  }

  void entry();
};

void
Ldr_names_parser::entry()
{
  Names::Name new_name = namedup(cur().text());
  Ldr::Entry *e = new Entry(new_name, Names::Obj());
  int err = _ns->insert(e);
  if (err)
    {
      freename(new_name);
      delete e;
      throw Parse_exception(_scan, cur(), "name already exists:");
    }
  cap_rval(e);
}

void
read_namespaces(Ldr::Script *cfg)
{
  L4::Cap<void> ns_ns_cap = Gate_alloc::alloc(&ns_ns);
  L4Re::Env::env()->names()->register_obj("ns", ns_ns_cap, L4Re::Namespace::Rw | L4Re::Namespace::Dir);
  cfg->rewind_stream();
  cxx::String name;
  while (cfg->goto_section("namespace", &name))
    {
      if (name.empty())
	{
	  Err().printf("%s:%d: cannot create anonymous name space, skip\n",
	               cfg->file(), cfg->line() - 1);
	  continue;
	}

      Names::Entry *n = ns_ns.find(name);

      if (n)
	{
	  Err().printf("%s:%d: object with name '%.*s' already exists, skip.\n",
	               cfg->file(), cfg->line() - 1, name.len(), name.start());
	  continue;
	}

      Name_space *ns = new Name_space();
      if (!ns)
	{
	  Err().printf("%s:%d: cannot allocate name space, out of memory\n",
	               cfg->file(), cfg->line() - 1);
	  continue;
	}

      enum { Ns_flags = Names::Obj::F_dir | Names::Obj::F_rw };
      L4::Cap<void> ns_cap = Gate_alloc::alloc(ns);

      info.printf("added name space '%.*s'\n", name.len(), name.start());

      Ldr::Scanner s2 = cfg->block(1);
      Ldr_names_parser nparse(&s2, ns);
      nparse.parse(Err());
      if (nparse.errors())
	{
	  Err().printf("=================================================\n");
	  Err().printf("could not create name space: %.*s\n", name.len(), name.start());
	  Err().printf("=================================================\n");
	  // XXX: fix memory leak we should free the name space (ns)
	  continue;
	}

      ns_ns.register_obj(name, Names::Obj(Ns_flags, ns));

      if (info.is_active())
	{
	  info.printf("dump of user-define name space: '%.*s'\n",
	              name.len(), name.start());
	  ns->dump();
	}
    }

  if (Dbg(Dbg::Boot).is_active())
    {
      info.printf("dump of user defined name spaces\n");
      ns_ns.dump(true);
    }

}

}
