/*
 * (c) 2008-2009 Technische Universität Dresden
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */
#include "loader.h"
#include "debug.h"
#include "obj_reg.h"

#include <l4/libloader/script>
#include <l4/cxx/auto_ptr>
#include <l4/cxx/iostream.h>
#include <l4/cxx/l4iostream.h>
#include <l4/cxx/exceptions>
#include <l4/cxx/ipc_helper>

#include <l4/sys/scheduler>
#include <l4/sys/utcb.h>
#include <l4/sys/kernel_object.h>
#include <l4/sys/factory>
#include <l4/sys/task>
#include <l4/sys/thread>

#include "name_space.h"
#include "config_namespace.h"
#include "script_invoke.h"

#include <l4/re/l4aux.h>
#include <l4/re/env>
#include <l4/re/service>

#include <cctype>
#include <cstring>

using L4Re::Rm;
using L4Re::Dataspace;
using L4Re::Env;
using L4Re::chksys;
using L4Re::Util::Auto_cap;

using L4::Cap;
static Dbg info(Dbg::Info);

namespace Names { using namespace L4Re::Util::Names; }

extern l4re_aux_t* l4re_aux;

Ldr::Script::Bit_desc _scr_redbg[] =
{{"info",       Dbg::Info},
 {"nfo",        Dbg::Info},
 {"warn",       Dbg::Warn},
 {"boot",       Dbg::Boot},
 {"server",     Dbg::Server},
 {"svr",        Dbg::Server},
 {"exceptions", Dbg::Exceptions},
 {"exc",        Dbg::Exceptions},
 {"POSIX",      Dbg::Cmd_line},
 {"ldso",       Dbg::Cmd_line},
 {"loader",     Dbg::Loader},
 {"ldr",        Dbg::Loader},
 {"namespace",  Dbg::Name_space},
 {"ns",         Dbg::Name_space},
 {"all",        ~0UL},
 {0,0}};

Ldr::Script::Bit_desc _scr_ldr_flags[] =
{{"eager_map",    L4RE_AUX_LDR_FLAG_EAGER_MAP},
 {"all_segs_cow", L4RE_AUX_LDR_FLAG_ALL_SEGS_COW},
 {"pinned_segs",  L4RE_AUX_LDR_FLAG_PINNED_SEGS},
 {0,0}};

enum
{
  Stack_size        = 0x8000,
  Stack_address     = 0xb0000000,
  Kip_address       = 0xa0000000,
};


static
Cap<Dataspace>
__alloc_app_stack(App_task *a, Loader::Info *ti, unsigned long size)
{
  Auto_cap<Dataspace>::Cap stack = L4Re::Util::cap_alloc.alloc<Dataspace>();

  if (!stack.is_valid())
    chksys(-L4_ENOMEM, "cannot allocate capability");

  chksys(a->allocator()->alloc(size, stack.get()),
         "cannot allocate stack data space");

  l4_addr_t stack_local = 0;
  chksys(L4Re::Env::env()->rm()->attach(&stack_local, size,
                                        Rm::Search_addr, stack.get(), 0));

  ti->stack.set_local_top((char*)(stack_local + size));
  return stack.release();
}


class Ldr_exec_parser : public Ldr::Parser
{
private:
  Ldr_obj_mgr mgr;
public:
  Value_parser_set prog_opts;
  Ldr::Capability_parser ns_cap;
  Ldr::Capability_parser log_cap;
  Ldr::Capability_parser factory_cap;
  Ldr::Capability_parser mem_cap;
  Ldr::Capability_parser scheduler_cap;
  Ldr::Num_or_enum_parser<unsigned> l4re_dbg;
  Ldr::Num_or_enum_parser<unsigned> ldr_flags;
  Ldr::Number_parser<unsigned> base_prio, max_prio, max_threads;
  Ldr::Number_parser<l4_umword_t> cpu_affinity;

  explicit Ldr_exec_parser(Ldr::Scanner *scan);

  void entry();
};

Ldr_exec_parser::Ldr_exec_parser(Ldr::Scanner *scan)
: Ldr::Parser(scan, &mgr),
  ns_cap("ns"), log_cap("log"), factory_cap("factory"), mem_cap("mem"),
  scheduler_cap("scheduler"), l4re_dbg("l4re_dbg", _scr_redbg, 0),
  ldr_flags("ldr_flags", _scr_ldr_flags, 0),
  base_prio("base_prio"), max_prio("max_prio"), max_threads("max_threads"),
  cpu_affinity("affinity")
{
  prog_opts.add(&l4re_dbg);
  prog_opts.add(&ldr_flags);
  prog_opts.add(&base_prio);
  prog_opts.add(&max_prio);
  prog_opts.add(&ns_cap);
  prog_opts.add(&log_cap);
  prog_opts.add(&factory_cap);
  prog_opts.add(&mem_cap);
  prog_opts.add(&scheduler_cap);
  prog_opts.add(&max_threads);
  prog_opts.add(&cpu_affinity);
}


void
Ldr_exec_parser::entry()
{
  Ldr::Token lval = cur();
  check_assignment();

  shift();
  if (!prog_opts.parse(this, lval))
    throw Ldr::Parse_exception(_scan, lval, "unexpected option:");

  if (cur().type() != Ldr::Token::Semicolon)
    throw Ldr::Parse_exception(_scan, cur(), "expected ';', got:");
}

// Pilfered from MOE -- start
static
bool operator == (cxx::String const &a, char const *b)
{
  if ((int)strlen(b) != a.len())
    return false;

  return strncmp(a.start(), b, a.len()) == 0;
}

static
void
read_global_config(Ldr::Script *cfg)
{
  cfg->rewind_stream();
  if (!cfg->goto_section("config"))
    return;

  class Cfg_parser : public Ldr::Parser
  {
  public:
    unsigned dbg_lvl;
    explicit Cfg_parser(Ldr::Scanner *scan) : Ldr::Parser(scan, 0) {}
    Val find_object(Ldr::Token const &name)
    {
      throw Ldr::Parse_exception(_scan, name, "object not found:");
    }
    void entry()
    {
      Ldr::Token lval = cur();
      check_assignment();
      shift();
      if (lval.text() == "dbg")
	{
	  if (cur().type() == Ldr::Token::Number)
	    dbg_lvl = number();
	  else
	    dbg_lvl |= enum_list((Ldr::Script::Bit_desc const *)_scr_redbg);
	}
      else
	throw Ldr::Parse_exception(_scan, lval, "unexpected option:");

      if (cur().type() != Ldr::Token::Semicolon)
	throw Ldr::Parse_exception(_scan, cur(), "expected ';', got:");
    }

    Val invoke(Val const &, Ldr::Token const &func)
    { throw Ldr::Parse_exception(_scan, func, "no invocation allowed"); }
  };

  Ldr::Scanner scan = cfg->block(1);
  Cfg_parser parser(&scan);
  parser.dbg_lvl = 0;
  parser.parse(Err());

  Dbg::set_level(parser.dbg_lvl);
}
// Pilfered from MOE -- end

bool Loader::start(char const *cmdline)
{
  //Dbg ldr(Dbg::Loader);

  L4Re::Env const *e = L4Re::Env::env();

  info.printf("Loading: %s\n", cmdline);

  unsigned long cmdline_len = strlen(cmdline);

  Auto_cap<Dataspace>::Cap file = L4Re::Util::cap_alloc.alloc<Dataspace>();
  if (!file.is_valid())
    chksys(-L4_ENOMEM, "cannot allocate capability");

  chksys(e->names()->query(cmdline, file.get()), cmdline);

  if (!check_file_type(file.get()))
    {
      unsigned long size = file->size();
      char const *data = 0;
      chksys(e->rm()->attach(&data, l4_round_page(size),
                             Rm::Search_addr | Rm::Read_only, file.get(), 0));
      Ldr::Script scr(data, size, cmdline, Dbg(Dbg::Parser, "scr"), Err());

      read_global_config(&scr);
      Ldr::read_namespaces(&scr);

      scr.rewind_stream();

      while (scr.next_entry())
	exec(&scr);

      file.release();
      //chksys(e->rm()->detach(l4_addr_t(data), 0));
      return true;
    }

  Ldr::Script src(cmdline, cmdline_len, "<builtin>", Dbg(Dbg::Parser, "scr"), Err());

  return exec(&src);
}

void
inline
throw_user_exception(long err)
{
  throw(L4::Runtime_error(err));
}

void
inline
check(l4_msgtag_t const &t)
{
  if (t.has_error())
    L4::throw_ipc_exception(Cap<void>::Invalid, t, l4_utcb());

  if (t.label() < 0)
    throw_user_exception(t.label());
}

void
inline
check(int t)
{
  if (t < 0)
    throw_user_exception(t);
}

class Ldr_task : public App_task, public Names::Chained_observer
{
private:
  l4_umword_t _entry, _stack_ptr;
  Ldr::Script scriptlet;
  Loader *loader;

public:
  Ldr_task(Loader *l, Ldr::Script const &s)
  : _entry(0), _stack_ptr(0), scriptlet(s), loader(l)
  {}

  void set_entry_data(l4_umword_t entry, l4_umword_t stack_ptr)
  { _stack_ptr = stack_ptr; _entry = entry; }

  void notify()
  {
    loader->launch(this, &scriptlet);
  }
};

bool
Loader::exec(Ldr::Script *scr)
{
  cxx::Auto_ptr<Ldr_task> app_task(new Ldr_task(this, *scr));

  if (!app_task)
    {
      Err().printf("could not allocate task control block\n");
      return false;
    }

  Ldr::Scanner scan = Ldr::Scanner(*scr).block(1, 1);
  Ldr_exec_parser parser(&scan);
  parser.no_invokes();
  parser.parse(Err());

  scr->skip_entry();

  if (parser.errors())
    {
      Err().printf("===================================================\n");
      Err().printf("Could not parse entry\n");
      Err().printf("===================================================\n");
      return false;
    }

  L4::Cap<L4Re::Namespace> ns_cap = L4Re::Env::env()->names();
  L4Re::Util::Names::Name_space *ns = 0;
  parser.set_cap_if_valid(parser.ns_cap.data, &ns_cap, &ns);

  if (!ns || ns->is_complete())
    app_task->notify();
  else
    {
      info.printf("deferred startup, name space still incomplete\n");
      info.printf("waiting for");
      bool first = true;
      for (L4Re::Util::Names::Name_space::Const_iterator i = ns->begin(); i != ns->end(); ++i)
	if (i->is_placeholder())
	  {
	    info.cprintf("%s '%.*s'", first?"":",", i->name().len(), i->name().name());
	    first = false;
	  }
      info.cprintf(" to complete/appear\n");
      ns->add_observer(app_task.get());
    }

  app_task.release();
  return true;
}

bool
Loader::launch(App_task *app_task, Ldr::Script *scr)
{
  class SSi : public L4kproxy::Scheduler_interface
  {
  public:
    int info(l4_umword_t *, l4_sched_cpu_set_t *)
    { return -L4_ENOSYS; }

    int run_thread(L4::Cap<L4::Thread> thread, l4_sched_param_t const &sp)
    { return l4_error(L4Re::Env::env()->scheduler()->run_thread(thread, sp)); }

    int idle_time(l4_sched_cpu_set_t const &) { return -L4_ENOSYS; }
  };

  static SSi _default_scheduler;

  enum
  {
    Utcb_area_start        = 0xb3000000,
    Factory_cap            = 2,
    Rm_thread_cap          = 3,
    External_rm_cap        = 5,
    Allocator_cap          = 7,
    Names_cap,
    Log_cap,
    Parent_cap,
    Kip_cap,
    Scheduler_cap,
    First_free,
    Default_max_threads    = 16,
    Total_max_threads      = 256,
  };

  Dbg info(Dbg::Info);

  enum Prios
  {
    Default_thread_prio = 2,
    Default_base_prio = 0x00,
    Default_max_prio  = 0xff,
  };

  char const *const re_binary = "rom/l4re";
  Cap<Dataspace> l4re = L4Re::Util::cap_alloc.alloc<Dataspace>();
  if (!l4re.is_valid())
    chksys(-L4_ENOMEM, "cannot allocatoe capability");

  int err = Env::env()->names()->query(re_binary, l4re);

  if (err == -L4_ENOENT)
    {
      Err().printf("could not find '%s'!\n", re_binary);
      return false;
    }
  chksys(err, re_binary);

  Info task_info;

  // set default values for utcb area, values may be changed by loader
  task_info.utcbs_start = Utcb_area_start;
  task_info.utcbs_log2size  = l4util_log2(Default_max_threads * L4_UTCB_OFFSET);

  // set default values for the application stack
  task_info.stack.set_target_stack(Stack_address, Stack_size);
  task_info.kip = Kip_address;

  if (!read_infos(l4re, &task_info))
    {
      // XXX: have to free task ID here
      Err().printf("invalid binary '%s' while loading\n", re_binary);

      return false;
    }

  Cap<Dataspace> stack = __alloc_app_stack(app_task, &task_info, task_info.stack.stack_size());

#if 0
  //char const *env0 = "LD_DEBUG=detail move sym bind reloc";
  char const *env0 = "LD_DEBUG=";
  env0 = task_info.stack.push_object(env0, strlen(env0) + 1);
#else
  char const *env0 = "";
  env0 = task_info.stack.push_object(env0, strlen(env0) + 1);
#endif
  char const *argv0 = 0, *argvl = 0;
  scr->parse_entry(&task_info.stack, &argv0, &argvl);

  Ldr::Scanner scan = scr->block(1);
  Ldr_exec_parser parser(&scan);
  parser.l4re_dbg = 0;
  parser.base_prio = Default_base_prio;
  parser.max_prio  = Default_max_prio;
  parser.max_threads = Default_max_threads;
  parser.cpu_affinity = ~0UL;
  parser.parse(Err());
  if (parser.errors())
    {
      Err().printf("===================================================\n");
      Err().printf("Could not start: %s\n", argv0);
      Err().printf("===================================================\n");
      return false;
    }

  Cap<void> task_mem_alloc = Env::env()->mem_alloc();
  parser.set_cap_if_valid(parser.mem_cap.data, &task_mem_alloc);
  // might still happen because loading is only defered when the namespace
  // is incomplete but not when e.g. 'mem = ns/foo/blah' could not be
  // resolved when the namespace got/is complete, workaround is to add
  // 'blah = ns/foo/blah'
  if (!task_mem_alloc.is_valid())
    {
      Err().printf("'mem' setting invalid\n");
      return false;
    }

  bool ns_writable = false;
  L4::Cap<L4Re::Namespace> ns_cap = L4Re::Env::env()->names();
  L4Re::Util::Names::Name_space *ns = 0;
  if (parser.set_cap_if_valid(parser.ns_cap.data, &ns_cap, &ns))
    ns_writable = parser.ns_cap.data.flags() & Names::Obj::F_rw;

  Cap<void> log_cap = Env::env()->log();
  parser.set_cap_if_valid(parser.log_cap.data, &log_cap);
  if (!log_cap.is_valid())
    {
      Err().printf("'log' setting invalid\n");
      return false;
    }

  Cap<L4::Factory> task_factory = Env::env()->factory();
  parser.set_cap_if_valid(parser.factory_cap.data, &task_factory);
  if (!task_factory.is_valid())
    {
      Err().printf("'factory' setting invalid\n");
      return false;
    }

  L4::Cap<L4::Scheduler> scheduler_cap = Env::env()->scheduler();
  L4kproxy::Scheduler_interface *scheduler_if = &_default_scheduler;
  void *scheduler = 0;
  parser.set_cap_if_valid(parser.scheduler_cap.data, &scheduler_cap, &scheduler);

  info.printf("loading '%s'\n", argv0);

  if (parser.cpu_affinity.data != ~0UL || parser.base_prio.data != Default_base_prio
      || parser.max_prio.data != Default_max_prio)
    {
      info.printf("    base_prio = 0x%x max_prio = 0x%x\n",
                  parser.base_prio.data, parser.max_prio.data);
      app_task->_sched.set_prio(parser.base_prio.data, parser.max_prio.data);
      app_task->_sched.restrict_cpus(parser.cpu_affinity.data);
      Gate_alloc::alloc(&app_task->_sched);
      scheduler_cap = L4::cap_cast<L4::Scheduler>(app_task->_sched.obj_cap());
      scheduler_if = &app_task->_sched;
    }

  task_info.stack.align(sizeof(l4_umword_t));
  char const *before_vma = task_info.stack.ptr();


  if (!load(l4re, app_task, &task_info))
    {
      // XXX: have to free task ID here
      Err().printf("invalid binary '%s' while loading '%s'\n",
                   re_binary, argv0);

      return false;
    }

  if (parser.max_threads.data != Default_max_threads)
    {
      if (parser.max_threads.data > Total_max_threads)
	{
	  Err().printf("Maximum supported threads per task: 256.\n");
	  throw L4::Runtime_error(-L4_ENOMEM, "too many threads");
	}
      task_info.utcbs_log2size = l4util_log2(l4util_next_power2(parser.max_threads.data) * L4_UTCB_OFFSET);
      info.printf("    max_threads = %3d\n", parser.max_threads.data);
    }

  l4_addr_t utcb_area_start
    = task_info.stack.add
        (task_info.utcbs_start,
         1UL << task_info.utcbs_log2size,
         app_task,
         L4::Cap<Dataspace>::Invalid, 0, 0, true, task_info.utcbs_log2size);

  task_info.stack.add(task_info.stack.target_addr(),
      task_info.stack.stack_size(),
      app_task, stack, 0, false, true);
#if 0
  L4::cout << "LDR:   stack @ "
           << L4::n_hex(task_info.stack.target_ptr()) << '\n';
#endif

  //L4::cerr << "AUX: " << l4re_aux << "\n";
  task_info.stack.add(task_info.kip, L4_PAGESIZE, app_task,
      Cap<Dataspace>(l4re_aux->kip_ds), 0, true);


  task_info.stack.set_ma_cnt(before_vma);

  Cap<void> child_kip_cap(Kip_cap << L4_CAP_SHIFT);

  // ldr_flags for l4re
  task_info.stack.push(l4_umword_t(parser.ldr_flags.data));

  // debug level for l4re
  task_info.stack.push(l4_umword_t(parser.l4re_dbg.data));

  // KIP DS capability
  task_info.stack.push(l4_umword_t(child_kip_cap.cap()));

  void const **local_aux = task_info.stack.push_local_ptr(argv0);
  L4Re::Env *env = task_info.stack.push(L4Re::Env());

#if 0
  if (app_task->log.tag() == 0)
    {
      unsigned long l = strlen(argv0);
      unsigned long i;
      for (i = l; i > 0 && argv0[i] != '/'; --i)
	;

      if (argv0[i] == '/')
	++i;

      app_task->log.set_tag(argv0 + i, l - i);
    }
#endif
  Gate_alloc::alloc(app_task);
  // L4::cout << "TBC: " << app_task << ", " << app_task->obj_cap() << "\n";

  app_task->task_cap(L4Re::Util::cap_alloc.alloc<L4::Task>());
  app_task->thread_cap(L4Re::Util::cap_alloc.alloc<L4::Thread>());

  env->mem_alloc      (Cap<L4Re::Mem_alloc>(Allocator_cap << L4_CAP_SHIFT));
  env->parent         (Cap<L4Re::Parent>(Parent_cap << L4_CAP_SHIFT));
  env->scheduler      (Cap<L4::Scheduler>(Scheduler_cap << L4_CAP_SHIFT));
  env->rm             (Cap<L4Re::Rm>(External_rm_cap << L4_CAP_SHIFT));
  env->names          (Cap<L4Re::Namespace>(Names_cap << L4_CAP_SHIFT));
  env->log            (Cap<L4Re::Log>(Log_cap << L4_CAP_SHIFT));
  env->main_thread    (Cap<L4::Thread>(Rm_thread_cap << L4_CAP_SHIFT));
  env->factory        (Cap<L4::Factory>(Factory_cap << L4_CAP_SHIFT));
  env->first_free_cap (First_free);
  env->utcb_area      (l4_fpage(utcb_area_start, task_info.utcbs_log2size, 0));
  env->first_free_utcb(utcb_area_start + L4_UTCB_OFFSET);


  // AUXV_NULL
  task_info.stack.align(sizeof(l4_umword_t));
  task_info.stack.push(0UL);
  task_info.stack.push(0UL);

  task_info.stack.align(sizeof(l4_umword_t));
  task_info.stack.push_local_ptr(env);
  task_info.stack.push(l4_umword_t(0xf1));

  task_info.stack.push_local_ptr(local_aux);
  task_info.stack.push(l4_umword_t(0xf0));

  // Env terminator
  task_info.stack.push(0UL);
  task_info.stack.push_local_ptr(env0);


  // ARGV stuff
  l4_umword_t argc = 0;

  // ARGV terminator
  task_info.stack.push(0UL);

  // put arguments, if any, onto the target stack
  if (argv0)
    {
      do
	{
	  if (argvl == argv0)
	    {
	      // just use the basename for ARGV[0]
	      argvl = strrchr(argv0, '/');

	      if (!argvl)
		argvl = argv0;
	      else
		++argvl;
	    }
          task_info.stack.push_local_ptr(argvl);
	  ++argc;

	  // scan for previous argument, remember the stack is top down
	  for (; argvl < argv0 && *argvl; ++argvl)
	    ;

	  ++argvl;
	}
      while (argvl <= argv0);
    }

  // ARGC
  task_info.stack.push(argc);

  Cap<L4::Task> ntask = app_task->task_cap();
  Cap<L4::Thread> nthread = app_task->thread_cap();
  Cap<L4::Factory> factory(L4Re::Env::env()->factory());

  check(factory->create_task(ntask, env->utcb_area()));
  check(factory->create_thread(nthread));
  check(ntask->map(L4Re::This_task, ntask.fpage(),
                   Cap<L4::Task>(L4Re::This_task).snd_base()));
  check(ntask->map(L4Re::This_task, task_factory.fpage(),
                   env->factory().snd_base()));
  check(ntask->map(L4Re::This_task, nthread.fpage(),
                   env->main_thread().snd_base()));
  check(ntask->map(L4Re::This_task, scheduler_cap.fpage(),
                   env->scheduler().snd_base()));
  check(ntask->map(L4Re::This_task, app_task->obj_cap().fpage(),
                   env->parent().snd_base()));
#if 0
  check(ntask->map(L4Re::This_task,
                   ns->obj_cap().fpage(ns_writable?L4_FPAGE_RWX:L4_FPAGE_RW),
                   env->names().snd_base()));
#else
  check(ntask->map(L4Re::This_task,
                   ns_cap.fpage(ns_writable?L4_FPAGE_RWX:L4_FPAGE_RW),
                   env->names().snd_base()));
#endif
  check(ntask->map(L4Re::This_task, log_cap.fpage(),
                   env->log().snd_base()));
  check(ntask->map(L4Re::This_task, app_task->rm()->obj_cap().fpage(),
                   env->rm().snd_base()));
  check(ntask->map(L4Re::This_task, task_mem_alloc.fpage(),
                   env->mem_alloc().snd_base()));
  check(ntask->map(L4Re::This_task, Cap<void>(l4re_aux->kip_ds).fpage(),
                   child_kip_cap.snd_base()));

  L4::Thread::Attr th_attr;

  th_attr.pager(env->rm());
  th_attr.exc_handler(env->rm());
  th_attr.scheduler(env->rm());
  th_attr.bind((l4_utcb_t*)utcb_area_start, ntask);

  check(nthread->control(th_attr));
  l4_sched_param_t sp;
  sp.prio = Default_thread_prio;
  sp.quantum = 0;
  sp.affinity = l4_sched_cpu_set(0, ~0);

  check(scheduler_if->run_thread(nthread, sp));

  nthread->ex_regs(task_info.entry, task_info.stack.target_ptr(), 0);

  return true;
}
