/*
 * (c) 2008-2009 Technische Universität Dresden
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */
#include "loader_elf.h"
#include "app_task.h"

#include <l4/re/dataspace>
#include <l4/re/rm>
#include <l4/re/dataspace>
#include <l4/re/env>

#include "debug.h"

#include <cstring>
#include <l4/libloader/elf>
#include <l4/re/error_helper>
#include <l4/re/util/cap_alloc>

using L4::Cap;
using L4Re::Env;
using L4Re::Dataspace;
using L4Re::Rm;
using L4Re::chksys;
using L4Re::Util::cap_alloc;
using namespace Ldr;

static Dbg ldr(Dbg::Loader, "ldr");

struct Phdr_load
{
  l4_addr_t base;
  Cap<Dataspace> bin;
  App_task *task;
  Elf_loader::Info *info;

  Phdr_load(l4_addr_t base, Cap<Dataspace> const &bin,
            App_task *task, Elf_loader::Info *info)
  : base(base), bin(bin), task(task), info(info)
  {}

  void operator () (Elf_phdr const &ph) const
  {
    if (ph.type() != PT_LOAD)
      return;

    if (!ph.memsz())
      return;

    char *paddr = (char*)(l4_trunc_page(ph.paddr()) + base);
    l4_umword_t offs = l4_trunc_page(ph.offset());
    l4_umword_t page_offs = ph.offset() & (L4_PAGESIZE-1);
    l4_umword_t fsz  = ph.filesz();
    if (fsz && page_offs != (ph.paddr() & (L4_PAGESIZE-1)))
      {
	Err().printf("malformed ELF file, file offset and paddr mismatch\n");
	chksys(-L4_EINVAL, "malformed elf file");
      }

    l4_umword_t size = l4_round_page(ph.memsz() + page_offs);

    if ((ph.flags() & PF_W) || ph.memsz() > fsz)
      {
	// copy section
	Cap<Dataspace> mem = cap_alloc.alloc<Dataspace>();
	if (!mem.is_valid())
	  chksys(-L4_ENOMEM, "ELF loader: could not allocate capability");
	chksys(task->allocator()->alloc(size, mem), "loading writabel ELF segment");
	if (mem->copy_in(0, bin, offs, fsz + page_offs) < 0)
	  {
	    L4::Cap<L4Re::Rm> r = L4Re::Env::env()->rm();
	    L4Re::Rm::Auto_region<char const *> data;
	    chksys(r->attach(&data, fsz + page_offs,
	                     Rm::Search_addr | Rm::Read_only, bin, offs));

	    memcpy(paddr + page_offs, data.get() + page_offs, fsz);
	  }

	if (!info->stack.add(l4_addr_t(paddr), size, task, mem, 0))
	  {
	    ldr.printf("  ERROR: add vm region for segemnt\n");
	    chksys(-L4_ENOMEM, "could not attach ELF segment");
	  }
      }
    else
      {
	// map from file
	if (!info->stack.add(l4_addr_t(paddr), size, task, bin,
	      offs, true))
	  {
	    ldr.printf("  ERROR: add vm region for segemnt\n");
	    chksys(-L4_ENOMEM, "could not attach ELF segment");
	  }
      }

  }
};

bool
Elf_loader::read_infos(Cap<Dataspace> bin, Info *info)
{
  Cap<Rm> rm = L4Re::Env::env()->rm();
  L4Re::Rm::Auto_region<Elf_ehdr const *> eh;

  chksys(rm->attach(&eh, L4_PAGESIZE, Rm::Search_addr | Rm::Read_only, bin, 0),
         "attaching ELF file");

  if (!eh->is_valid())
    {
      Err().printf("file C[%lx] is not an ELF binary\n", bin.cap());
      chksys(-L4_EINVAL, "not an ELF binary");
    }


  Phdr_stack_kip stack_kip(info->stack.target_addr(), info->stack.stack_size(),
      info->kip);

  eh->iterate_phdr(stack_kip);

  info->kip = stack_kip.kip;
  info->stack.set_target_stack(stack_kip.stack_addr, stack_kip.stack_size);

  ldr.printf("  PHDRs: type  offset\tpaddr\tvaddr\tfilesz\tmemsz\trights\n");
  eh->iterate_phdr(Phdr_print<Dbg>(ldr));

  return true;
}

bool
Elf_loader::load(Cap<Dataspace> bin, App_task *task, Info *info)
{
  Cap<Rm> rm = L4Re::Env::env()->rm();
  L4Re::Rm::Auto_region<Elf_ehdr const *> eh;

  chksys(rm->attach(&eh, L4_PAGESIZE, Rm::Search_addr | Rm::Read_only, bin, 0),
         "attaching ELF file");

  if (!eh->is_valid())
    {
      Err().printf("file C[%lx] is not an ELF binary\n", bin.cap());
      chksys(-L4_EINVAL, "not an ELF binary");
    }


  eh->iterate_phdr(Phdr_load(0, bin, task, info));

  info->entry = eh->entry();

  ldr.printf(" done...\n");

  return true;
}

bool
Elf_loader::check_file_type(L4::Cap<L4Re::Dataspace> file) const
{
  L4Re::Rm::Auto_region<void *> bin;
  chksys(Env::env()->rm()->attach(&bin, L4_PAGESIZE,
                                  Rm::Search_addr | Rm::Read_only, file, 0));
  bool match = (memcmp(bin.get(), "\177ELF", 4) == 0);
  return match;
}
